# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Parse Git commit messages."""

# import logging
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional

from lark import Lark
from versioning.config import GRAMMAR_PATH


class CommitMessageParser:
    """Parse commit messages."""

    def __init__(
        self,
        grammar_path: str = GRAMMAR_PATH,
        start: str = 'message',
        **kwargs: Any
    ) -> None:
        """Initialize commit message parser."""
        # TODO: need to limit messages to 100 characters
        self.types = kwargs.pop('types', ['feat', 'fix'])
        self.scopes = kwargs.pop('scopes', [])
        self.__parser = Lark.open(grammar_path, start=start, **kwargs)

    def parse(
        self,
        text: str,
        start: Optional[str] = None,
        on_error: Optional[Callable] = None
    ) -> None:
        """Parse commit message."""
        self.__tree = self.__parser.parse(  # type: ignore
            text, start=start, on_error=on_error
        )

    def _get_section(self, name: str) -> Optional[Any]:
        """Get commit message section."""
        for arg in self.__tree.children:
            # NOTE: will not have parse tree for non-match
            if (
                hasattr(arg, '__dict__')
                and vars(arg)['data'] == name
            ):
                return arg
        return None

    @property
    def title(self) -> Dict[str, Any]:
        """Get title section of commit message."""
        title: Dict[str, Any] = defaultdict(lambda: None)
        section = self._get_section('title')
        if section:
            for arg in section.children:
                title[arg.data] = next((x.value for x in arg.children), True)
        return title

    @property
    def body(self) -> List[str]:
        """Get body section of commit message."""
        section = self._get_section('body')
        if section:
            return [arg.value for arg in section.children]
        else:
            return []

    @property
    def footer(self) -> Dict[str, Any]:
        """Get footer section of commit message."""
        footer: Dict[str, Any] = defaultdict(lambda: None)
        footer['issues'] = []

        section = self._get_section('footer')
        if section:
            for arg in section.children:
                if arg.data == 'trailer':
                    footer['trailer'] = {}
                    for x, y in enumerate(arg.children):
                        if x == 0:
                            footer['trailer']['token'] = y.value
                        if x == 1:
                            footer['trailer']['name'] = y.value
                        if x == 2:
                            footer['trailer']['email'] = y.value
                if arg.data == 'issue':
                    footer['issues'].append(
                        {arg.children[0].value: arg.children[1].value}
                    )
                if arg.data == 'breaking_change':
                    footer['breaking_change'] = next(
                        (x.value for x in arg.children),
                        'Unknown'
                    )
        return footer



# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Manage version numbers."""

# import logging
from typing import Any, List, Optional, Tuple

from packaging.version import Version as PackageVersion
from packaging.version import _cmpkey, _parse_local_version, _Version
from transitions import Machine


class Version(PackageVersion):
    """Provide PEP440 compliant versioning."""

    def __init__(self, version: str, **kwargs: Any) -> None:
        """Initialize version object."""
        # self.kind = kwargs.pop('version_system', 'semver')
        self.compat = kwargs.pop('compat', 'pep440')
        print(self.compat)
        super().__init__(version=version)

        # TODO: transitions here should be populated by VCS workflow
        self.enable_devreleases = kwargs.get('enable_devreleases', True)
        self.enable_prereleases = kwargs.get('enable_prereleases', True)
        self.enable_postreleases = kwargs.get('enable_postreleases', True)

        self.machine = Machine(
            self, states=self.states, initial=self.release_type
        )

        # dev-releases
        devrelease_sources = ['final']
        if self.enable_postreleases:
            devrelease_sources.append('post')
        self.machine.add_transition(
            trigger='start_devrelease',
            source=devrelease_sources,
            dest='development',
            before='new_devrelease',
            conditions=['enable_devreleases'],
        )

        # pre-releases
        prerelease_sources = []
        if self.enable_devreleases:
            prerelease_sources.append('development')
        else:
            prerelease_sources.append('final')
            if self.enable_postreleases:
                prerelease_sources.append('post')
        self.machine.add_transition(
            trigger='start_alpha',
            source=prerelease_sources,
            dest='alpha',
            before='new_prerelease',
            conditions=['enable_prereleases'],
        )
        self.machine.add_transition(
            trigger='start_beta',
            source='alpha',
            dest='beta',
            before='new_prerelease',
            conditions=['enable_prereleases'],
        )
        self.machine.add_transition(
            trigger='start_release',
            source='beta',
            dest='release',
            before='new_prerelease',
            conditions=['enable_prereleases'],
        )

        # final release
        final_sources = []
        if self.enable_prereleases:
            final_sources.append('release')
        elif self.enable_devreleases:
            final_sources.append('development')
        else:
            final_sources.append('final')
            if self.enable_postreleases:
                final_sources.append('post')
        self.machine.add_transition(
            trigger='finish_release',
            source=final_sources,
            dest='final' if 'final' not in final_sources else None,
            before='finalize_release',
        )

        # post-releases
        self.machine.add_transition(
            trigger='start_postrelease',
            source='final',
            dest='post',
            before='new_postrelease',
            conditions=['enable_postreleases'],
        )

        self.machine.add_transition(
            trigger='bump_release',
            source='*',
            dest=None,
            before='_bump_release',
        )

    def __str__(self) -> str:
        """Handle version formatting."""
        parts = []

        # Epoch
        if self.epoch != 0:
            parts.append(f"{self.epoch}!")

        # Release segment
        parts.append('.'.join(str(x) for x in self.release))

        # Pre-release
        if self.pre is not None:
            if self.compat == 'semver':
                separator = '-'
            else:
                separator = ''
            parts.append(
                f"{separator}{self.pre[0]}{str(self.pre[1])}"
            )

        # Post-release
        if self.post is not None:
            if self.compat == 'semver':
                separator = '-'
            else:
                separator = '.post'
            parts.append(f"{separator}{self.post}")

        # Development release
        if self.dev is not None:
            if self.compat == 'semver':
                separator = '-'
            else:
                separator = '.'
            parts.append(f"{separator}dev{self.dev}")

        # Local version segment
        if self.local is not None:
            parts.append(f"+{self.local}")

        return ''.join(parts)

    @property
    def states(self) -> List[str]:
        """List all states."""
        states = ['final']
        if self.enable_devreleases:
            states += ['development']
        if self.enable_prereleases:
            states += ['alpha', 'beta', 'release']
        if self.enable_postreleases:
            states += ['post']
        return states

    @property
    def release_type(self) -> str:
        """Get the current state of package release."""
        if self.is_devrelease:
            state = 'development'
        elif self.is_prerelease and self.pre:
            if self.pre[0] == 'a':
                return 'alpha'
            elif self.pre[0] == 'b':
                return 'beta'
            elif self.pre[0] == 'rc':
                return 'release'
        elif self.is_postrelease:
            state = 'post'
        else:
            state = 'final'
        return state

    def __update_version(
        self,
        epoch: Optional[int] = None,
        release: Optional[Tuple[Any, ...]] = None,
        pre: Optional[Tuple[str, Optional[int]]] = None,
        post: Optional[Tuple[Optional[str], Optional[int]]] = None,
        dev: Optional[Tuple[str, Optional[int]]] = None,
        local: Optional[str] = None,
    ) -> None:
        """Update the internal version state."""
        if not (epoch or release):
            pre = pre or self.pre
            post = post or (None if self.post is None else ('post', self.post))
            dev = dev or (None if self.dev is None else ('dev', self.dev))

        self._version = _Version(
            epoch=epoch or self.epoch,
            release=release or self.release,
            pre=pre,
            post=post,
            dev=dev,
            local=_parse_local_version(local) if local else None,
        )

        self._key = _cmpkey(
            self._version.epoch,
            self._version.release,
            self._version.pre,
            self._version.post,
            self._version.dev,
            self._version.local,
        )

    def bump_epoch(self) -> None:
        """Update epoch releaes for version system changes."""
        self.__update_version(epoch=self.epoch + 1)
        self.machine.set_state('final')

    def bump_major(self) -> None:
        """Update major release to next version number."""
        self.__update_version(release=(self.major + 1, 0, 0))
        self.machine.set_state('final')

    def bump_minor(self) -> None:
        """Update minor release to next version number."""
        self.__update_version(release=(self.major, self.minor + 1, 0))
        self.machine.set_state('final')

    def bump_micro(self) -> None:
        """Update micro release to next version number."""
        self.__update_version(release=(self.major, self.minor, self.micro + 1))

    def __bump_version(self, kind: str) -> None:
        """Bump version based on version kind."""
        if kind == 'major':
            self.bump_major()
        if kind == 'minor':
            self.bump_minor()
        if kind == 'micro':
            self.bump_micro()

    def _bump_release(self, *args: Any, **kwargs: Any) -> None:
        """Update to the next development release version number."""
        if self.enable_devreleases and self.dev is not None:
            dev = self.dev + 1
            self.__update_version(dev=('dev', dev))
        if self.enable_prereleases and self.pre is not None:
            pre = (self.pre[0], self.pre[1] + 1)
            self.__update_version(pre=pre)
        if self.enable_postreleases:
            if self.release_type == 'final':
                self.__update_version(post=('post', 0))
            elif self.post is not None:
                post = self.post + 1
                self.__update_version(post=('post', post))

    def new_devrelease(self, kind: str = 'minor') -> None:
        """Update to the next development release version number."""
        if self.dev is None:
            self.__bump_version(kind)
            self.__update_version(dev=('dev', 0))

    def new_prerelease(self, kind: Optional[str] = None) -> None:
        """Update to next prerelease version type."""
        if self.pre is not None:
            if self.pre[0] == 'a':
                pre = ('b', 0)
            elif self.pre[0] == 'b':
                pre = ('rc', 0)
        else:
            if kind:
                self.__bump_version(kind)
            else:
                self.finalize_release()
            pre = ('a', 0)
        self.__update_version(pre=pre)

    def new_postrelease(self) -> None:
        """Update the post release version number."""
        if self.release_type == 'final':
            self.__update_version(post=('post', 0))

    def new_local(self, build: str = 'build-0') -> None:
        """Create new local version instance number."""
        self.__update_version(local=build)

    def finalize_release(self) -> None:
        """Update to next prerelease version type."""
        if self.is_devrelease or self.is_prerelease:
            self.__update_version(release=(self.major, self.minor, self.micro))



# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Provide exceptions."""


class PromanVersioningException(Exception):
    """Provide base errors in PackageManager."""



# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Parse git commit messages."""

# import logging
import os
from typing import Any, List, Optional

from pygit2 import GIT_OBJ_COMMIT, Commit, Repository, Signature, Tag

from versioning.exception import PromanVersioningException


class Git:
    """Provide settings for git repositories."""

    def __init__(self, repo: Repository) -> None:
        """Initialize git object."""
        self.repo = repo
        self.hooks_dir = os.path.join(self.repo.path, 'hooks')
        self.config = os.path.join(self.repo.path, 'config')

    @property
    def repo_dir(self) -> str:
        """Return directory of repository."""
        return self.repo.path

    @property
    def working_dir(self) -> str:
        """Return working directory of project."""
        return os.path.abspath(os.path.join(self.repo_dir, '..'))

    @property
    def branch(self) -> str:
        """Retrieve the current branch."""
        return self.repo.head.name

    def commit(
        self,
        branch: Optional[str] = None,
        filepaths: List[str] = [],
        **kwargs: Any,
    ) -> Optional[Commit]:
        """Create commit."""
        ref = f"refs/heads/{branch}" if branch else self.branch

        user = list(self.repo.config.get_multivar('user.name'))
        if user == []:
            raise PromanVersioningException('git user.name not configured')
        email = list(self.repo.config.get_multivar('user.email'))
        if email == []:
            raise PromanVersioningException('git user.email not configured')

        author = Signature(user[-1], email[-1])
        committer = kwargs.get('commiter', author)
        message = kwargs.get('message', 'ci: bump version')

        # populate index
        index = self.repo.index
        if filepaths == []:
            index.add_all()
        else:
            for filepath in filepaths:
                index.add(
                    os.path.relpath(
                        filepath, os.path.join(self.repo.path, '..')
                    )
                )
        index.write()
        tree = index.write_tree()

        # get parent
        parents = kwargs.get('parents', [self.repo.head.target])
        encoding = kwargs.get('encoding', 'utf-8')

        # commit
        if not kwargs.get('dry_run', False):
            commit = self.repo.create_commit(
                ref,
                author,
                committer,
                message,
                tree,
                parents,
                encoding,
            )
            return commit
        # TODO: should a mock commit be created?
        return None

    def tag(
        self, name: str, branch: Optional[str] = None, **kwargs: Any
    ) -> Optional[Tag]:
        """Create tag."""
        ref = f"refs/heads/{branch}" if branch else self.branch
        commit = self.repo.resolve_refish(ref)[0]
        oid = commit.hex
        kind = kwargs.get('kind', GIT_OBJ_COMMIT)

        # XXX: regression from GitPython, signature not available here
        signature = kwargs.get('signature', commit.signature)
        message = kwargs.get('message', f"ci: {name}")
        if not kwargs.get('dry_run', False):
            tag = self.repo.create_tag(name, oid, kind, signature, message)
            return tag
        # TODO: should a mock tag be created?
        return None



# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Convenience tools to manage project versioning."""

import logging
from typing import Any, List

from pygit2 import Repository

from versioning.config import CONFIG_FILES, REPO_DIR, Config
from versioning.controller import IntegrationController
from versioning.exception import PromanVersioningException
from versioning.vcs import Git
from versioning.version import Version  # noqa

__author__ = 'Jesse P. Johnson'
__author_email__ = 'jpj6652@gmail.com'
__title__ = 'proman-versioning'
__description__ = 'Convenience tool to manage project versioning.'
__version__ = '0.3.0a0'
__license__ = 'MPL-2.0'
__copyright__ = 'Copyright 2021 Jesse Johnson.'
__all__ = (
    'IntegrationController',
    'get_source_tree',
    'get_release_controller',
    'Version'
)

logging.getLogger(__name__).addHandler(logging.NullHandler())


def get_source_tree(config_files: List[str] = CONFIG_FILES) -> Config:
    """Get source tree from path."""
    try:
        config = Config(filepaths=config_files)
        return config
    except Exception as err:
        raise PromanVersioningException(err)


def get_release_controller(*args: Any, **kwargs: Any) -> IntegrationController:
    """Create and return a release controller."""
    repo_dir = kwargs.pop('repo_dir', REPO_DIR)
    repo = Git(Repository(repo_dir))

    config_files = kwargs.pop('config_files', CONFIG_FILES)
    cfg = get_source_tree(config_files=config_files)

    return IntegrationController(config=cfg, repo=repo, **kwargs)



# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Provide CLI management."""

from argufy import Parser

from versioning import __version__
from versioning import cli


def main() -> None:
    """Provide CLI for git-tools."""
    parser = Parser(version=__version__)
    parser.add_commands(cli)
    parser.dispatch()


if __name__ == '__main__':
    main()



# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Control project versions."""

import logging
import sys
from typing import Optional

from versioning import get_release_controller

_log = logging.getLogger(__name__)
_controller = get_release_controller()


def bump(
    commit: bool = False,
    release: bool = False,
    # tag: bool = False,
    # tag_name: Optional[str] = None,
    message: Optional[str] = None,
    # sign: bool = False,
    build: Optional[str] = None,
    dry_run: bool = False,
) -> None:
    """Manage project versions.

    Parameters
    ----------
    commit: bool
        Commit release changes to project.
    message: str
        Notes included with the commit.
    release: bool
        Make this version a new release.
    build: str
        Build version from other systems the may help coordinate build
        artifacts.
    dry_run: bool
        Run the command without making changes.

    """
    # tag: bool
    #     Tag this version.
    # tag_name: str
    #     The tag name to use for this tag.
    #     default: version
    # sign: bool
    #     Sign commit with PKI signature.
    version = _controller.bump_version(
        commit=commit,
        release=release,
        # tag=tag,
        # tag_name=tag_name,
        message=message,
        # sign_tag=sign,
        build=build,
        dry_run=dry_run,
    )
    print(str(version), file=sys.stdout)


def view(release: bool = False, filepaths: bool = False) -> None:
    """Get the current version or settings of a project.

    Parameters
    ----------
    release: bool
        View the current release type
    filepaths: bool
        List the version filepaths

    """
    if release:
        print(_controller.release, file=sys.stdout)
    if filepaths:
        for x in _controller.config.templates:
            print(x['filepath'], file=sys.stdout)
    if not release and not filepaths:
        print(_controller.config.version, file=sys.stdout)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['versioning', 'versioning.grammars']

package_data = \
{'': ['*'], 'versioning': ['templates/*']}

install_requires = \
['argufy>=0.1.1-alpha.12,<0.2.0',
 'cffi>=1.15.0,<2.0.0',
 'compendium>=0.1.1-beta.0,<0.2.0',
 'lark-parser>=0.10.0,<0.11.0',
 'packaging>=21.3,<22.0',
 'pygit2>=1.6.1,<2.0.0',
 'transitions>=0.8.4,<0.9.0']

entry_points = \
{'console_scripts': ['version = versioning.__main__:main']}

setup_kwargs = {
    'name': 'proman-versioning',
    'version': '0.3.0a0',
    'description': 'Project Manager Versioning tool.',
    'long_description': '# Proman Versioning\n\n[![License](https://img.shields.io/badge/License-MPL%202.0-blue.svg)](https://spdx.org/licenses/MPL-2.0)\n[![Build Status](https://travis-ci.org/kuwv/proman-versioning.svg?branch=master)](https://travis-ci.org/kuwv/proman-versioning)\n[![codecov](https://codecov.io/gh/kuwv/proman-versioning/branch/master/graph/badge.svg)](https://codecov.io/gh/kuwv/proman-versioning)\n\n## Overview\n\nProject Manager Versioning is a PEP-440 compliant tool for automating project\nversions using conventional commits.\n\n## Install\n\n`pip install proman-versioning`\n\n## Setup\n\nThis tool is designed to work with any textfile using a templating pattern and\n path to the file.\n\n### Configuring versions\n\nConfiguration can be performed with either the `.versioning` or `pyproject.toml`\nfiles.\n\n#### Global configuration settings:\n\nSpecific types of releases can be disabled by setting the respective release to\nfalse.\n\nDisable development releases:\n```\nenable_devreleases = false\n```\n\nDisable pre-releases:\n```\nenable_prereleases = false\n```\n\nDisable post-releases:\n```\nenable_postreleases = false\n```\n\n#### File specific settings:\n\nUse different version compatibiliy type:\n```\ncompat = "semver"\n```\n\nUpdate only the release version for a configuration:\n```\nrelease_only = true\n```\n\n#### Example `.versioning`\n\nThe `.versioning` config is a non-specfile based project file using TOML. This is the\npreferred configuration for non-python projects that may use this tool.\n\n```\n[proman]\nversion = "1.2.3"\n\n[proman.versioning]\ndisable_devreleases = true\n\n[[tool.proman.versioning.files]]\nfilepath = "pyproject.toml"\npattern = "version = \\"${version}\\""\n\n[[proman.versioning.files]]\nfilepath = "example/__init__.py"\npattern = "__version__ = \'${version}\'"\n```\n\n#### Example `pyproject.toml`\n\n```\n[tool.proman]\nversion = "1.2.3"\n\n[tool.proman.versioning]\n\n[[tool.proman.versioning.files]]\nfilepath = "pyproject.toml"\npattern = "version = \\"${version}\\""\n\n[[tool.proman.versioning.files]]\nfilepath = "example/__init__.py"\npattern = "__version__ = \'${version}\'"\n```\n\n#### Example `setup.cfg`\n\nSetuptools allows `setup.cfg` to pull the version from the application. This\nshould be used in tandem with either of the above configurations to control\nversions for a project.\n\n```\n[metadata]\nname = example\nversion = attr: src.VERSION\n...\n```\n\n## References\n\n- https://www.conventionalcommits.org/en/v1.0.0/\n- https://www.python.org/dev/peps/pep-0440/\n- https://semver.org\n- https://calver.org\n',
    'author': 'Jesse P. Johnson',
    'author_email': 'jpj6652@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Provide configuration for versioning tools."""

import os
from dataclasses import InitVar, dataclass, field
from typing import Any, Dict, List

from compendium.config_manager import ConfigManager
from pygit2 import discover_repository

from versioning.exception import PromanVersioningException
from versioning.version import Version

# from urllib.parse import urljoin, urlparse

# TODO check VCS for paths
CURRENT_DIR = os.getcwd()
REPO_DIR = discover_repository(CURRENT_DIR)
if REPO_DIR is None:
    raise PromanVersioningException('Unable to locate git repository.')
PROJECT_DIR = os.path.abspath(os.path.join(REPO_DIR, '..'))
CONFIG_FILES = [
    os.path.join(PROJECT_DIR, '.versioning'),
    os.path.join(PROJECT_DIR, 'pyproject.toml'),
    # TODO: add setup.cfg
]

GRAMMAR_PATH: str = os.path.join(
    os.path.dirname(__file__), 'grammars', 'conventional_commits.lark'
)
# 'versioning/templates/gitmessage.j2'


# @dataclass
# class GitConfig:
#     """Manage git config."""
#
#     system_config: str = os.path.join(os.sep, 'etc', 'gitconfig')
#     global_config: str = os.path.join(os.path.expanduser('~'), '.gitconfig')


@dataclass
class ParserConfig:
    """Configure parser operation."""

    config: InitVar[Dict[str, Any]] = None
    types: List[str] = field(default_factory=list)
    scopes: List[str] = field(default_factory=list)

    def __post_init__(self, config: Dict[str, Any]) -> None:
        """Configure VCS message parsing."""
        # thinking builtin types might not need to be here
        # if (
        #     self.types == []
        #     and 'types' in config
        #     and config['types'] != []
        # ):
        #     self.types = config['types']
        #     if 'feat' not in self.types:
        #         self.types.insert(0, 'feat')
        #     if 'fix' not in self.types:
        #         self.types.insert(1, 'fix')
        #     if 'release' not in self.types:
        #         self.types.insert(2, 'release')
        # else:
        #     self.types = ['feat', 'fix', 'release']

        if (
            self.types == []
            and config is not None
            and 'types' in config
            and config['types'] != []
        ):
            self.types = config['types']

        if (
            self.scopes == []
            and config is not None
            and 'scopes' in config
            and config['scopes'] != []
        ):
            self.scopes = config['scopes']


@dataclass
class ReleaseConfig:
    """Configure release operation."""

    config: InitVar[Dict[str, Any]] = None
    enable_devreleases: bool = True
    enable_prereleases: bool = True
    enable_postreleases: bool = True

    def __post_init(self, config: Dict[str, Any]) -> None:
        """Load configuration for release operation."""
        if 'enable_devreleases' in config:
            self.enable_devreleases = config['enable_devreleases']
        if 'enable_prereleases' in config:
            self.enable_prereleases = config['enable_prereleases']
        if 'enable_postreleases' in config:
            self.enable_postreleases = config['enable_postreleases']


# @dataclass
# class TemplateConfig:
#     """Configure template layout."""
#
#     filepath: str
#     pattern: str
#     release_only: bool = False


@dataclass
class Config(ConfigManager):
    """Manage settings from configuration file."""

    filepaths: InitVar[List[str]]
    defaults: InitVar[Dict[str, Any]] = field(default={})
    templates: List[Dict[str, Any]] = field(default_factory=list)
    parser: ParserConfig = field(init=False)
    release: ReleaseConfig = field(init=False)
    version: Version = field(init=False)

    def __post_init__(
        self, filepaths: List[str], defaults: Dict[str, Any]
    ) -> None:
        """Initialize settings from configuration."""
        # XXX: config_manager is not passing separator
        super().__init__(filepaths=filepaths, separator='.', defaults=defaults)

        # load configuration from paths in order or precedence
        config = self.lookup(
            '.proman.versioning',
            '.tool.proman.versioning',
            '.tool.poetry.versioning',
        ) or {}

        if (
            self.templates == []
            and 'files' in config
            and config['files'] != []
        ):
            self.templates = config['files']
        else:
            raise PromanVersioningException('no configuration files provided')

        if 'types' not in config:
            angular_convention = [
                'build',
                'ci',
                'docs',
                'perf',
                'refactor',
                'style',
                'test',
            ]
            config['types'] = angular_convention

        self.parser = ParserConfig(config=config)
        self.release = ReleaseConfig(config=config)

        config_version = self.lookup(
            '.proman.version',
            '.tool.proman.version',
            '.tool.poetry.version',
        )
        if config_version is None:
            raise PromanVersioningException('no version found in filepaths')

        self.version = Version(
            version=config_version,
            enable_devreleases=self.release.enable_devreleases,
            enable_prereleases=self.release.enable_prereleases,
            enable_postreleases=self.release.enable_postreleases,
            compat=config.get('compat', 'pep440'),
        )



# copyright: (c) 2021 by Jesse Johnson.
# license: MPL-2.0, see LICENSE for more details.
"""Coordinate actions from commit messages."""

import logging
import os
import re
import sys
from copy import deepcopy
from string import Template
from typing import TYPE_CHECKING, Any, Dict

from versioning.exception import PromanVersioningException
from versioning.grammars.conventional_commits import CommitMessageParser
from versioning.version import Version

if TYPE_CHECKING:
    from versioning.config import Config
    from versioning.vcs import Git

# from transitions import Machine

# TODO: version comparison against previous version
# has API spec been modified?
# has Python version changed?
# has requirements versions changed?

# TODO: build numbers
# git revision
# GNU build ID
# timestamp
# external (ex: CI/CD build number)

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


# TODO determine relation with state and git hooks
class IntegrationController(CommitMessageParser):
    """Control version releases."""

    # workflow_types = ['rolling', 'sustainment']
    # Trunk Based Development (TBD) - GitHub flow
    # Stage Based Development (SBD) - DTAP
    # Release Branching Strategy (RBS) - PEP440
    # Feature Branching Strategy (FBS) - Agile

    def __init__(
        self,
        config: 'Config',
        repo: 'Git',
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize commit message action object."""
        self.config = config
        # parse_current_branch = kwargs.pop('parse_current_branch', True)
        message = kwargs.pop('message', None)
        super().__init__(*args, **kwargs)

        self.vcs = repo
        if message is None:
            head = self.vcs.repo.head
            target = self.vcs.repo[head.target]
            self.parse(target.message)
            log.debug(f"provided commit message: '{message}'")
        else:
            self.parse(message)
            log.debug(f"found commit message: '{message}'")

    @property
    def release(self) -> str:
        """Get the current version release state."""
        return self.config.version.state  # type: ignore

    @staticmethod
    def __get_version_regex(version: Version) -> str:
        """Get PEP-440 compliant regex for version."""
        # TODO: should this be part of Version?
        r = '.'.join(str(x) for x in version.release)
        if version.dev:
            v = f"{r}[-_\\.]?dev[-_\\.]?{version.dev or '0?'}"
            return v

        if version.pre:
            if version.epoch > 0:
                v = f"{version.epoch}!{r}"
            pre = version.pre[0]
            inst = version.pre[1]
            if pre == 'a' or pre == 'alpha':
                v = f"{r}[-_\\.]?(?:a|alpha)[-_\\.]?{inst or '0?'}"
            if pre == 'b' or pre == 'beta':
                v = f"{r}[-_\\.]?(?:b|beta)[-_\\.]?{inst or '0?'}"
            if pre == 'rc' or pre == 'release':
                v = f"{r}[-_\\.]?(?:rc|release)[-_\\.]?{inst or '0?'}"
            return v
        elif version.post:
            v = f"{r}[-_\\.]?(?:post[-_\\.]?)?{version.post}"
            return v
        else:
            return str(version)

    def __update_config(
        self,
        config: Dict[str, Any],
        new_version: Version,
        dry_run: bool = False,
    ) -> None:
        """Update target file with new version."""
        if 'release_only' in config and config['release_only']:
            current_release = '.'.join(
                str(x) for x in self.config.version.release
            )
            version = Version(current_release)

            release = '.'.join(
                str(x) for x in new_version.release
            )
            new_version = Version(release)
        else:
            version = deepcopy(self.config.version)

        if 'compat' in config:
            version.compat = config['compat']
            new_version.compat = config['compat']

        pattern = config['pattern']
        filepath = os.path.join(
            self.vcs.working_dir, config['filepath']
        )

        # TODO: handle when file does not exist
        with open(filepath, 'r+') as f:
            file_contents = f.read()

            # prepare source expression match pattern
            version_str = IntegrationController.__get_version_regex(
                version=version
            )
            template = Template(pattern).substitute(version=version_str)
            match = re.compile(
                # XXX: escape not compiling correctly
                template,  # re.escape(template),
                flags=0,
            )
            log.debug(f"using pattern for source version {match}")

            # substitute the expression in file
            file_contents = match.sub(
                Template(pattern).substitute(version=str(new_version)),
                file_contents,
            )

            if not dry_run:
                # save the file
                try:
                    f.seek(0)
                    f.truncate()
                    f.write(file_contents)
                    log.info(f"writting file at: '{filepath}'")
                except Exception as err:
                    print(err, file=sys.stderr)
            else:
                # print the file changes
                print(file_contents, file=sys.stdout)
                log.info(
                    f"dry-run skipping file write at: '{filepath}'"
                )

    def update_configs(self, new_version: Version, **kwargs: Any) -> None:
        """Update version within config files."""
        dry_run = kwargs.pop('dry_run', False)
        stats = self.vcs.repo.diff('HEAD').stats
        if stats.files_changed == 0 or dry_run:
            if self.config.version != new_version:
                for config in self.config.templates:
                    self.__update_config(
                        config=config,
                        new_version=new_version,
                        dry_run=dry_run,
                    )

                self.config.version = new_version
                if kwargs.pop('commit', True):
                    # TODO: tie scope to configs, releases or version ranges
                    scope = 'version'
                    self.vcs.commit(
                        filepaths=[
                            f['filepath']
                            for f in self.config.templates
                        ],
                        message=(
                            f"ci({scope}): apply {str(new_version)} updates"
                        ),
                        dry_run=dry_run,
                    )
                    log.info(f"commiting version changes: {str(new_version)}")

                if kwargs.get('tag', False):
                    self.vcs.tag(
                        name=str(new_version),
                        ref='HEAD',
                        message=None,
                        dry_run=dry_run,
                    )
                    log.info(f"applying tag: {str(new_version)}")
            else:
                raise PromanVersioningException(
                    'no version update could be determined'
                )
        else:
            raise PromanVersioningException('repository is not clean')

    def start_release(self, **kwargs: Any) -> Version:
        """Start a release."""
        new_version = deepcopy(self.config.version)

        if (
            self.config.version.enable_devreleases
            and self.release == 'development'
        ):
            log.info('found devrelease')
            new_version.start_alpha()  # type: ignore

        if self.config.version.enable_prereleases:
            log.info('found prerelease')
            if self.release == 'alpha':
                new_version.start_beta()  # type: ignore
            if self.release == 'beta':
                new_version.start_release()  # type: ignore
            if self.release == 'release':
                new_version.finish_release()  # type: ignore

        if (
            self.release == 'final'
            or (
                self.config.version.enable_postreleases
                and self.release == 'post'
            )
        ):
            log.info(f"found {self.release} release")
            new_version.start_devrelease()  # type: ignore

        return new_version

    def bump_version(self, **kwargs: Any) -> Version:
        """Update the version of the project."""
        if (
            ('type' in self.title and self.title['type'] == 'release')
            or kwargs.get('release') is True
        ):
            new_version = self.start_release(**kwargs)
            self.update_configs(new_version, **kwargs)
        else:
            new_version = deepcopy(self.config.version)
            build = kwargs.pop('build', None)

            # TODO: break, and feat should start devrelease from final or post
            # local number depends on metadata / fork / conflict existing vers
            if self.title['break'] or self.footer['breaking_change']:
                new_version.bump_major()
            elif 'type' in self.title:
                if self.title['type'] == 'feat':
                    new_version.bump_minor()
                elif self.title['type'] == 'fix':
                    new_version.bump_micro()
                elif self.title['type'] in self.config.parser.types:
                    # new_version = self.__bump_release(new_version)
                    new_version.bump_release()  # type: ignore
                else:
                    # TODO: need debug statement here instead
                    raise PromanVersioningException(
                        'received unsupported commit type'
                    )

            # TODO: configure local version handling
            if build is not None:
                new_version.new_local(build=build)
            self.update_configs(new_version, **kwargs)
        return new_version
